#include <stdio.h>
#include <time.h>

#include "../../dprint.h"
#include "../../parser/msg_parser.h"
#include "../tm/t_funcs.h"
#include "../../error.h"
#include "../../ut.h"      /* q_memchr */

#include "fcp_mod.h"
#include "fcp.h"
#include "text_ops.h"

/* 
   Replaces Header in sip_msg. It will replace the header pointed
   by <key> with the header pointed by <str>. Both are NULL
   terminated strings
*/

int replace_header(struct sip_msg* msg, char* key, char* str, char *where)
{
  struct lump* l;
  regmatch_t pmatch;
  char* s;
  int len;
 
  int    status=-1;

  if(!where) return -1;
 
  /* Searches for <key> from the beginning of msg */

  printf("Replace_header: searching for:%s\nIts length: %d\n",
	 key, strlen(key));

  status = match(where, key, &pmatch);
  if(status!=0) {
    printf("Replace_header. Regexec returned %d\n", status);
    return -1;
  }

  printf("replace_header: Offset start= %d, offset end= %d\n", pmatch.rm_so, pmatch.rm_eo);

  if (pmatch.rm_so!=-1){
      
    if ((l=del_lump(&msg->add_rm, pmatch.rm_so,
		    pmatch.rm_eo-pmatch.rm_so, 0))==0) {
      printf("Error deleting lump\n");
      return -1;
    }
    len=strlen(str);
    s=pkg_malloc(len);
    if (s==0){
      LOG(L_ERR, "ERROR: replace_header: mem. allocation failure\n");
      return -1;
    }
    memcpy(s, str, len); 
    return insert_new_lump_after(l, s, strlen(str), 0)?1:-1;
  }

  return 1;

}


/* Replaces Contact header of SIP message */
 
int replace_contact(struct sip_msg *msg, struct fcp_info *fcp_tr) {
 
  str old;
  char *buf, *p;
  int len;

  old = msg->contact->body;

  /* Build the new Contact field */   

  len = /* 9, "Contact: " + */ 4 /* "sip:" */ + \
    fcp_tr->rpl_host.len + \
    1 /*":"*/ + fcp_tr->rpl_rtp_port.len /* port length */ + 2/*"<>"*/ \
    /* + CRLF_LEN */ + 1; /* for "\0"? */

  printf("Allocating memory for %d bytes\n", len);  

  buf = pkg_malloc( len );
  if(!buf) {
    LOG(L_ERR,"ERROR:replace_contact: out of memory! \n");
    return -1;
  }
  
  p = buf;
  //  append_str( p, "Contact: " , 9);
  *(p++) = '<';
  append_str( p, "sip:", 4);
  append_str( p, fcp_tr->rpl_host.s, fcp_tr->rpl_host.len);
  *(p++) = ':';
  append_str( p, fcp_tr->rpl_rtp_port.s, fcp_tr->rpl_rtp_port.len);
  *(p++) = '>';
  //  append_str( p, CRLF, CRLF_LEN);
  *(p++) = '\0';  

  printf("New header: %s\n", buf);

  if(replace_header(msg, old.s, buf, msg->orig)==-1) {
    printf("Could not replace header\n");
    return -1;
  }

  return 1;
}

/**************************************** 
   Replaces IP addresses and ports in SDP
   message with the info provided by an
   FCP server. Returns number of occurences.

   Does not work well. Only replaces PORT values.
   IP are not replaced. Why??
*****************************************/

int replace_sdp(struct sip_msg *msg, struct fcp_info *fcp_tr) {

  int len, count;
  int offset;
  char *ptr, *buf, *buf2, *buf3, *buf4, *string_end;
  
  count = 0;


  /* 
     Search for "IP4 <host>" and "audio <port>" strings 
     within the sip message. Replace their values accordingly
  */

  len = 4 /* "IP4 " */ + strlen(fcp_tr->sent_host.s);
  buf = pkg_malloc( len+1 ); /* extra character for \0? */
  if(!buf) {
    LOG(L_ERR,"ERROR:extcmd_add_contact: out of memory! \n");
    return -1;
  }
  
  ptr = buf;
  append_str( ptr, "IP4 " , 4);
  append_str( ptr, fcp_tr->sent_host.s, strlen(fcp_tr->sent_host.s));
  *(ptr) = '\0';

  len = 4 /* "IP4 " */ + strlen(fcp_tr->rpl_host.s);
  buf2 = pkg_malloc( len+1 ); /* extra character for \0? */
  if(!buf2) {
    LOG(L_ERR,"ERROR:replace_sdp: out of memory! \n");
    return -1;
  }
  
  ptr = buf2;
  append_str( ptr, "IP4 " , 4);
  append_str( ptr, fcp_tr->rpl_host.s, strlen(fcp_tr->rpl_host.s));
  *(ptr) = '\0';

  offset=0;
  string_end = msg->orig;

  //  while ((offset=get_offset_start(string_end, "IP4"))> 0){
    printf("Replace_SDP: Replacing %s with %s\n", buf, buf2);

    //    string_end = string_end + offset;
    
    if(replace_header(msg, buf, buf2, string_end)==-1) {
      printf("Replace_SDP. Could not replace IP\n");
      goto cont;
    }
    //    string_end=skip_delims(string_end);
    //    string_end=skip_chars(string_end);
    count++;
    //  }    



 cont:

  len = 6 /* "audio " */ + fcp_tr->rpl_rtp_port.len;
  buf3 = pkg_malloc( len+1 ); /* extra character for \0? */
  if(!buf3) {
    LOG(L_ERR,"ERROR:replace_sdp: out of memory! \n");
    return -1;
  }
  
  ptr = buf3;
  append_str( ptr, "audio " , 6);
  append_str( ptr, fcp_tr->sent_rtp_port.s, fcp_tr->sent_rtp_port.len);
  *(ptr) = '\0';

  len = 6 /* "audio " */ + sizeof(int);
  buf4 = pkg_malloc( len+1 ); /* extra character for \0? */
  if(!buf4) {
    LOG(L_ERR,"ERROR:replace_sdp: out of memory! \n");
    return -1;
  }
  
  ptr = buf4;
  append_str( ptr, "audio " , 6);
  append_str( ptr, fcp_tr->rpl_rtp_port.s, fcp_tr->rpl_rtp_port.len);
  *(ptr) = '\0';


  //  offset=0;
  //  string_end = msg->orig;

  //  while ((offset=get_offset_end(string_end, "audio"))> 0){

    printf("Replace_SDP: Replacing %s with %s\n", buf3, buf4);

    //    string_end = string_end + offset;

    if(replace_header(msg, buf3, buf4, string_end)==-1) {
      printf("Could not replace media port\n");
      goto error;
    }
    count++;
    
    //  }

 error:
    if(buf) pkg_free(buf);
    if(buf2) pkg_free(buf2);
    if(buf3) pkg_free(buf3);
    if(buf4) pkg_free(buf4);

    return count; 

}

/***************************************************

   Adds "received" and "rport" values to the via header

Example:

Via: SIP/2.0/UDP 10.1.1.1:4540;received=68.44.20.1;rport=9988

In parser/parse_via.h added PARAM_RPORT:

enum {
	PARAM_HIDDEN=230, PARAM_TTL, PARAM_BRANCH, 
	PARAM_MADDR, PARAM_RECEIVED, GEN_PARAM,
	PARAM_ERROR,  ********PARAM_RPORT*******
};


struct via_param {
	int type;
	str name;
	str value;
	int size;
	struct via_param* next;
};


struct via_body { 
	int error;
	str hdr;
	str name;
	str version;   
	str transport;
	str host;
	int port;
	str port_str;
	str params;
	str comment;
	int bsize;
	struct via_param* param_lst;
	struct via_param* last_param;

	struct via_param* branch;
	str tid; 
	struct via_param* received;
        
	******** struct via_param* rport; ********
	
	struct via_body* next;

};

****************************************************/

int replace_via(struct sip_msg *msg, struct fcp_info *fcp_tr) {


  char *host1, *port1;
  char *params1;
  int params1_len;
  char *hdr1;
  char *via_param1;
  int via_param1_len;


  //  str old;
  int via_len;
  char *buf, *p;
  int len;

  if(msg->via1) {

    printf("replace_sdp: Via present\n");

    host1= msg->via1->host.s;
    port1= msg->via1->port_str.s;
    if (msg->via1->params.len>0) {
      params1= msg->via1->params.s;
      params1_len = msg->via1->params.len;
      if (msg->via1->param_lst->value.len>0) {
	via_param1 = msg->via1->param_lst->value.s;
	via_param1_len = msg->via1->param_lst->value.len;
	printf("Via Param: %s. Size: %d\n", via_param1, via_param1_len);
      } else {
	printf("replace_sdp: No param_lst!\n");
	return -1;
      }
    } else {
      printf("replace_sdp: No via1->params!\n");
      return -1;
    }
    hdr1= msg->via1->hdr.s;

    via_len = msg->via1->bsize;


    /* Build the new Via using "received" and "rport" values */   

    len = via_param1_len /* Previous params */+ 1 \
      /*";"*/ + 9 /*"received=" + */ + \
      fcp_tr->rpl_host.len + \
      1 /*";"*/ + \
      6 /*"rport=" + */ + \
      fcp_tr->rpl_rtp_port.len + \
      /* 1 ";" + */ 1; /* for "\0"? */

  
    buf = pkg_malloc( len );
    if(!buf) {
      LOG(L_ERR,"ERROR:extcmd_add_contact: out of memory! \n");
      return -1;
    }
  
    p = buf;
    append_str( p, via_param1 , via_param1_len);
    append_str( p, ";" , 1);
    append_str( p, "received=" , 9);
    append_str( p, fcp_tr->rpl_host.s, fcp_tr->rpl_host.len);
    *(p++) = ';';
    append_str( p, "rport=" , 6);
    append_str( p, fcp_tr->rpl_rtp_port.s, fcp_tr->rpl_rtp_port.len);
    //    *(p++) = ';';


    *(p++) = '\0';  

    printf("New header: %s\n", buf);

    /*
    if(replace_header(msg, params1, buf, msg->orig)==-1) {
      printf("Could not insert "received" param\n");
      return -1;
    }
    msg->via1->params.len=strlen(buf);
    */

    /*
    if(insert_new_lump_after(msg->via1->rport->value.s, buf, strlen(buf), 0)==-1) {
      printf("Could not insert "rport" param\n");
      return -1;
    }
    */

   }

    /*
  if(msg->via2) {
    host2= msg->via2->host.s;
    port2= msg->via2->port_str.s;
    params2= msg->via2->params.s;
    hdr2= msg->via2->hdr.s;

    printf("Replace Via: Host2 = %s, Port2 = %s,Params2 = %s, HDR = %s\n",
	   host2, port2, params2, hdr2);

  }

    */

  return 1;
}

/* Not working properly */

int add_contact(struct sip_msg* msg , str* to_uri)
{
  struct lump_rpl *lump;
  char *buf, *p;
  int len;
  
  len = 9 /*"Contact: "*/ + to_uri->len + 2/*"<>"*/ + CRLF_LEN;
  
  buf = pkg_malloc( len );
  if(!buf) {
    LOG(L_ERR,"ERROR:extcmd_add_contact: out of memory! \n");
    return -1;
  }
  
  p = buf;
  append_str( p, "Contact: " , 9);
  *(p++) = '<';
  append_str( p, to_uri->s, to_uri->len);
  *(p++) = '>';
  append_str( p, CRLF, CRLF_LEN);
  
  lump = build_lump_rpl( buf , len );
  if(!lump) {
    LOG(L_ERR,"ERROR:extcmd_add_contact: unable to build lump_rpl! \n");
    pkg_free( buf );
    return -1;
  }
  add_lump_rpl( msg , lump );
  
  pkg_free(buf);
  return 1;
}
