/*
 * Firewall Control Protocol module
 *
 * $Id: acc_mod.c,v 1.11 2002/11/11 20:23:06 kozlik Exp $
 *
 * Copyright (C) 2001-2003 Fhg Fokus
 *
 * This file is part of ser, a free SIP server.
 *
 * ser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * ser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>

#include "../../sr_module.h"
#include "../../dprint.h"
#include "../../mem/mem.h"
#include "../tm/t_hooks.h"
#include "../tm/tm_load.h"
#include "../tm/h_table.h"

#include "../../parser/msg_parser.h"
#include "../../parser/parse_content.h"
#include "../../parser/parse_from.h"
#include "../../parser/parse_via.h"
#include "../../str.h"


#include "fcp_mod.h"
#include "fcp.h"
#include "fcp_pa.h"
#include "tcpClient.h"
#include "text_ops.h"
#include "parser.h"

struct tm_binds tmb;

static int mod_init( void );
static void destroy(void);
static int child_init(int rank);

db_con_t* db_handle;   /* Database connection handle. For future use? */

/* ----- Parameter variables ----------- */

char *fcp_server="127.0.0.1";
int fcp_port=12345;
int fcp_timer=60;
int fcp_flag=12345;
int usesyslog=1;
/* Use nat functionality? 0 No, 1 Yes */
int usenat=0;
/* IP address of SIP proxy */
char *sip_proxy_ip="192.168.6.153";
/* External IP address of NAT/FW */
char *firewall_ip="192.168.0.100";

/* flag which is needed for reporting missed calls */
int missed_flag=2;

/* report e2e ACKs too */
int report_ack = 1;

/* should early media replies (183) be logged ? default==no */
int early_media = 0;

/* should failed replies (>=3xx) be logged ? default==no */
int failed_transactions = 0;

/* ------------- Callback handlers --------------- */

static void fcp_onoutboundresp( struct cell* t,  struct sip_msg *msg,
	int code, void *param );
static void fcp_onack( struct cell* t,  struct sip_msg *msg,
	int code, void *param );
static void fcp_onoutbound( struct cell* t, struct sip_msg *msg,
	int code, void *param ) ;
static void fcp_onmissed(struct cell *t, struct sip_msg *reply,
	int code, void *param );
static void fcp_onend(struct cell *t, struct sip_msg *reply,
	int code, void *param );


struct module_exports exports= {
	"fcp",

	/* exported functions */
       	( char*[] ) { NULL /* "replace_contact" */},
	( cmd_function[] ) { NULL /* fcp_report */ },
	( int[] ) { 1/* What is that? */ },
	( fixup_function[]) { 0 /* what is that? */ },
	0, /* 1, number of exported functions */

	/* exported variables */
	(char *[]) { /* variable names */
		"fcp_server",
		"fcp_port",
		"fcp_timer",
		"fcp_flag",
		"usesyslog",
		"usenat",
		"sip_proxy_ip",
		"firewall_ip"
	},

	(modparam_t[]) { /* variable types */
		STR_PARAM,
		INT_PARAM,
		INT_PARAM,
		INT_PARAM,
		INT_PARAM,
		INT_PARAM,
		STR_PARAM,
		STR_PARAM
	},

	(void *[]) { /* variable pointers */
		&fcp_server,
		&fcp_port,
		&fcp_timer,
		&fcp_flag,
		&usesyslog,
		&usenat,
		&sip_proxy_ip,
		&firewall_ip
	},

	8,          /* number of variables */

	mod_init, 	/* initialization module */
	0,			/* response function */
    destroy,    /* destroy function */
	0,			/* oncancel function */
    child_init  /* per-child init function */
};



static int mod_init( void )
{

	load_tm_f	load_tm;

	fprintf( stderr, "fcp - initializing\n");

	/* import the TM auto-loading function */
	if ( !(load_tm=(load_tm_f)find_export("load_tm", NO_SCRIPT))) {
		LOG(L_ERR, "ERROR: fcp: mod_init: can't import load_tm\n");
		return -1;
	}
	/* let the auto-loading function load all TM stuff */
	if (load_tm( &tmb )==-1) return -1;

	/* register callbacks */
	if (tmb.register_tmcb( TMCB_REPLY, fcp_onoutboundresp, 0 /* empty param */ ) <= 0)
		return -1;
	if (tmb.register_tmcb( TMCB_E2EACK, fcp_onack, 0 /* empty param */ ) <=0 )
		return -1;
	if (tmb.register_tmcb( TMCB_REQUEST_OUT, fcp_onoutbound, 0 /* empty param */ ) <=0 )
		return -1;
	if (tmb.register_tmcb( TMCB_ON_NEGATIVE, fcp_onmissed, 0 /* empty param */ ) <=0 )
		return -1;
	//       	if (tmb.register_tmcb( TMCB_END, fcp_onend, 0 /* empty param */ ) <=0 )
	//  return -1;

	return 0;
}

static int child_init(int rank)
{

  sd = openConnection(fcp_server, fcp_port);

  if(sd<0) {
    perror("cannot send data ");
    closeConnection(sd);
    return -1;
  }

  return 0;

}

static void destroy(void)
{
#ifdef SQL_FCP
    if (db_handle) db_close(db_handle);
#endif

    /* Close the connection to the fcp server */
    closeConnection(sd);
}

static void fcp_onoutbound( struct cell* t, struct sip_msg *msg,
	int code, void *param )
{
	/* disable C timer for accounted calls */
  /*	if (isflagset( msg, fcp_flag)==1 ||
				(t->is_invite && isflagset( msg, missed_flag))) {
#		ifdef EXTRA_DEBUG
		DBG("DEBUG: noisy_timer set for firewall transversal\n");
#		endif
		t->noisy_ctimer=1;
	}
  */


  /* 
     An outbound request came. We do not know if the transaction 
     will complete.Therefore, in the case of only firewall, flag 
     the transaction for the time being.
     For NAT's, we must ask the NAT device what is the external IP 
     address and port that will be open. Then change the Via, Contact 
     fields in the SIP header, and the IP and media field in the SDP 
     header. Anything else? Then send the message to the destination.
  */

  int rc;
  struct media_info media;
  struct fcp_info fcp_message;
  char buffer[512];


  //  str sdp;
  //struct to_body  *from;
  //  struct via_body  *via;
  //str src[8];
  //  struct to_body  *to;

  /* Sanity check */

  if (!check_transaction_quadruple( msg )) {
    LOG(L_ERR,"ERROR:fcp_onoutbound bad message!\n");
    goto error;
  }


  /*********************************** 
     PRELIMINARY TESTS 
  ************************************/

  /* 
     Extract Via
  */

  /* My way to access the Via Header. Does not work :( */

  /*  struct to_body *via;
  
  if(!msg->h_via1) {
    LOG(L_ERR,"ERROR:fcp_onoutbound cannot get VIA header!\n");
    goto error;  
  }

  via = (struct to_body*)msg->h_via1->parsed;
  */

  /* First way to Extract Via */

  /*
    if(!msg->via1) {
    LOG(L_ERR,"ERROR: fcp_onoutbound cannot get VIA header!\n");
    goto error;
  }

  via = (struct via_body*)msg->via1;

  */

  /* The proper way to access the VIA header? */

  /*

  struct via_body  *via;

  via = pkg_malloc(sizeof(struct via_body));
  if (via == 0) {
    LOG(L_ERR, "ERROR:fcp_onoutbound: out of pkg_memory\n");
    goto error;
  }
  memset(via, 0, sizeof(struct to_body));

  parse_via(msg->via1->host.s,msg->via1->host.s+msg->via1->host.len+1,via);

  if (via->error == PARSE_ERROR) {  
    LOG(L_ERR,"ERROR: Error in VIA header!\n");
    pkg_free(via);
    goto error;
  } 
  */    

  /* Important parameters */

  /*
  src[0]= msg->from->body;
  src[1]= msg->to->body;
  src[2]= msg->callid->body;
  src[3]= msg->first_line.u.request.uri;
  src[4]= get_cseq( msg )->number;
  */

  /* topmost Via is part of transaction key as well ! */

  /*
  src[5]= msg->via1->host;
  src[6]= msg->via1->port_str;
  if (msg->via1->branch) {
    src[7]= msg->via1->branch->value;
    printf("REQUEST URI: %s\nFROM: %s\nTO: %s\nCALL-ID: %s\nCSEQ: %s\nVIA: %s:%s;tag=%s\n", src[3].s, src[0].s,src[1].s,src[2].s, src[4].s, src[5].s, src[6].s, src[7].s);
  } else {
    printf("REQUEST URI: %s\nFROM: %s\nTO: %s\nCALL-ID: %s\nCSEQ: %s\nVIA: %s:%s\n", src[3].s, src[0].s,src[1].s,src[2].s, src[4].s, src[5].s, src[6].s);
  }
  */

  /*  char headers[MD5_LEN];

  int test;
  test = char_msg_val(msg, headers);
  if (!test) {
    LOG(L_ERR,"ERROR: fcp_onoutbound cannot get headers!\n");
    goto error;    
  }
  */

   /* 
      Extract From parameters
   */ 

  /* if ( parse_from_header(msg)==-1 ) {
    LOG(L_ERR,"ERROR:fcp_onoutbound cannot get FROM header!\n");
    goto error;
  }
  from = (struct to_body*)msg->from->parsed;
  */

  /* Extract To header. In our case, not needed */

  
  /*  to = pkg_malloc(sizeof(struct to_body));
  if (to == 0) {
    LOG(L_ERR, "ERROR:fcp_onoutbound: out of pkg_memory\n");
    goto error;
  }
  memset(to, 0, sizeof(struct to_body));
  */

//  if (!msg->to) {
  
/*  parse_to(msg->to->body.s,msg->from->body.s+msg->from->body.len+1,to);

  if (to->error == PARSE_ERROR) {  
    LOG(L_ERR,"ERROR:fcp_onoutbound no TO header found!\n");
    pkg_free(to);
    goto error;
    } 
*/
//  to = (struct to_body*)msg->to->parsed;


  /* Print the results */

  //printf("VIA: %s\n", via->body.s); /* For "my way" */
  //  printf("VIA: %s:%d\n", via->host.s, via->port);
  //  printf("VIA: %s\n", via->hdr.s);
  //  printf("TO:%s\n", to->body.s);
  //printf("FROM: %s\n", from->body.s);
  //  printf("SDP BODY:\n%s\n", sdp.s);
  printf("\n%s\n", msg->orig);

  /* adds contact header into reply */
  /*  
  if (extcmd_add_contact(msg,&(get_to(msg)->uri))==-1) {
    LOG(L_ERR,"ERROR:extcmd:dump_msg: can't build contact for reply\n");
    goto error;
  }
  */

  /******************************** 
     END OF PREMILINARY TESTS 
  *********************************/


  /* Is it an outbound INVITE ? */

  if (t->is_invite) {

    if (usenat) {
      
      /* 
	 Test: replace Contact to point to NAT device
      */

      rc = parse_media_info(msg, msg, &media);

      if (rc<0) {
	printf("Error while extracting media params: \n");
	goto error;
      }


      rc = get_nat(sd, media.src_host.s , media.src_port, 
		   media.src_port+1);
      
      if(rc<0) {
	perror("Error sending request to NAT device");
	closeConnection(sd);
	goto error;
      }
      
      rc = read_reply(sd, buffer);
      
      if(rc<0) {
	perror("Error listening to connection ");
	closeConnection(sd);
	goto error;
      }
      
      /* Keep info related to the FCP transaction */
    
      rc = parse_fcp_info(buffer, &fcp_message, &media);
    
      if(rc<0) {
	perror("error listening to connection ");
	goto error;
      }

      rc = replace_contact(msg, &fcp_message);
      
      if(rc<0) {
	perror("Error replacing contact ");
	goto error;
      }

      /* 
	 Does not work well. Only port value
	 gets changed. Not IP address. Need to
	 change other fields, like via?
      */

      rc = replace_sdp(msg, &fcp_message);
      
      if(rc<0) {
	perror("Error replacing sdp ");
	goto error;
      }

      rc = replace_via(msg, &fcp_message);
      
      if(rc<0) {
	perror("Error replacing via ");
	goto error;
      }

    

    } /* endif use_nat */    

    /* setflag to recognise a flagged INVITE in subsequent responses */

    DBG("DEBUG: Invite. Setting fcp flag... \n");
    setflag(msg, fcp_flag);

  
  } /* Endif is_invite */

  else return;

error:
	//if (buf) pkg_free(buf);
  return;
}

static void fcp_onmissed(struct cell *t, struct sip_msg *reply,
	int code, void *param )
{
  //	struct sip_msg *rq;

  //	rq = t->uas.request;

  //	if (t->is_invite
  //			&& missed_flag
  //			&& isflagset( rq, missed_flag)==1
  //			&& code>=300 )
  //	{
  //		fcp_missed_report( t, reply, code);
		/* don't come here again on next failed branch */
  //		resetflag(rq, missed_flag );
  //	}

/* A negative response came, like a 400 (Not Found, Unauthorised...).
   We must close any pinhole associated with that transaction. If a 
   300 redirect is received, need to get the Contact field and initiate
   a new INVITE to the new address */
}

static void fcp_onoutboundresp( struct cell* t, struct sip_msg *reply,
	int code, void *param )
{

	struct sip_msg *rq;

	struct media_info media;
	struct fcp_info fcp_message;

	int res;

	int rc = 0;
	char buffer[512];

	/* Get the associated request */

	rq = t->uas.request;

	/* fcp_onreply is bound to TMCB_REPLY which may be called
	   we should not miss this event for missed calls either
	*/

//	fcp_onmissed(t, reply, code, param );

	/* if fcp enabled for flagged transaction, check if flag matches */

	if (fcp_flag && isflagset( rq, fcp_flag ))
	  printf("Transaction previously flagged!!!\n");
	/* early media is reported only if explicitely demanded,
	   other provisional responses are always dropped  */
//	if (code < 200 && ! (early_media && code==183))
//		return;
	/* negative transactions reported only if explicitely demanded */
//	if (!failed_transactions && code >=300) return;

	/* anything else, i.e., 2xx, always reported */
//	fcp_reply_report(t, reply, code);

/* 
   A 200 response came from the remote end. We must open the media 
   port contained in the SDP media header. Ignore the following 
   ACK from the local end point 
*/

	/* Sanity check */

	if (!check_transaction_quadruple( reply )) {
	  printf("fcponoutbound Bad Message!\n");
	  LOG(L_ERR,"ERROR:fcp_onoutbound bad message!\n");
	  goto error;
	}

	/* Print message for debugging purposes */

	printf("\n%s\n", reply->orig);

	/* Check if it is a 200 OK */
	if(!is_method(reply->orig, "200")) {
	  printf("Message is not 200 OK. Ignore\n");
	  goto error;	  
	}
	
	res = parse_media_info(rq, reply, &media);
	if (res<0) {
	  printf("Error while extracting media params: \n");
	  goto error;
	}
	
	/*
	  Open the firewall for the media to flow. If there is
	  nat, ask first to reserve ports
	*/

    if (usenat) {
      

      rc = get_nat(sd, media.src_host.s , media.src_port, 
		   media.src_port+1);

      if(rc<0) {
	perror("cannot send request to NAT device");
	closeConnection(sd);
	goto error;
      }

      rc = read_reply(sd, buffer);
  
      if(rc<0) {
	perror("error listening to connection ");
	closeConnection(sd);
	goto error;
      }

      /* Keep info related to the FCP transaction */

      rc = parse_fcp_info(buffer, &fcp_message, &media);
  
      if(rc<0) {
	perror("error listening to connection ");
	goto error;
      }

      /* 
	 With the info from the NAT, open the ports for
	 the media
      */
      
      res = open_pinhole(sd, media.src_host.s,
			 media.dst_host.s, media.src_port,
			 media.dst_port,
			 fcp_timer);

      rc = read_reply(sd, buffer);

      if(rc<0) {
	perror("error listening to connection ");
	closeConnection(sd);
	goto error;
      }

      /*
      rc = replace_contact(reply, &fcp_message);

      if(rc<0) {
	perror("error replacing contact ");
	printf("Error replacing contact \n");
	goto error;
      }
      */

    } /* endif(use_nat)*/

    /* 
       No NAT. Only open ports in firewall
    */
    else {

      res = open_pinhole(sd, media.src_host.s, media.dst_host.s,
			 media.src_port, media.dst_port, fcp_timer);
      
      rc = read_reply(sd, buffer);
      
      if(rc<0) {
	perror("error listening to connection ");
	closeConnection(sd);
	goto error;
      }
    }

error:
	//if (buf) pkg_free(buf);
  return;

}


static void fcp_onack( struct cell* t , struct sip_msg *ack,
	int code, void *param )
{
	struct sip_msg *rq;

	rq = t->uas.request;
	/* only for those guys who insist on seeing ACKs as well */
	//	if (!report_ack) return;
	/* if fcp enabled for flagged transaction, check if flag matches */
	//	if (fcp_flag && isflagset( rq, fcp_flag )==-1) return;

	//	fcp_ack_report(t, ack);
}

static void fcp_onend( struct cell* t, struct sip_msg *msg,
		       int code, void *param ) {
  printf("\n%s\n", msg->orig);

}
