#include "fcp_pa.h"
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>


/*
 * FCP Protocol Adaptor
 */


static int seq = 1234;
char buffer[2048];


int open_pinhole(int _s, char* _src, char* _dst, int _src_port, int _dst_port, int _t)
{
	int w, tmp;
	int len;

	if (!_src) return 0;
	if (!_dst) return 0;
	
	if (_t == 0) return 1;

	snprintf(buffer, 2048,
		 "SET FCP=1.0 SEQ=%d \n"   \
		 "PME PROTO=17; "     \
		 "SRCIP=%s; "      \
		 "DSTIP=%s; "	 \
		 "SRCPORT=%d-%d; "
		 "DSTPORT=%d-%d; \n" \
		 "SOPT ACTION=pass; "   \
		 "TIMER=%d \n"      \
		 "\n"              \
		 , seq++, _src, _dst, _src_port, _src_port + 1, _dst_port, _dst_port + 1, _t);


	len = strlen(buffer);
	w = 0;

	while(w != len) {
		tmp = write(_s, buffer + w, len - w);
		if (tmp == -1) {
			LOG(L_ERR, "open_pinhole(): %s\n", strerror(errno));
			return 0;
		}
		printf("Sent:\n%s", buffer);
			
		w += tmp;
	}
	DBG("open_pinhole(): %s \n", buffer);	
	DBG("open_pinhole(): %d bytes sent\n", w);

	return 1;
}


int close_pinhole(int _s, char* _src, char* _dst, int _src_port, int _dst_port)
{
	int w, tmp;
	int len;

	if (!_src) return 0;
	if (!_dst) return 0;
	
	snprintf(buffer, 2048,
		 "RELEASE FCP=1.0 SEQ=%d \n"   \
		 "PME PROTO=17; "     \
		 "SRCIP=%s; "      \
		 "DSTIP=%s; "	 \
		 "SRCPORT=%d-%d; "
		 "DSTPORT=%d-%d; \n" \
		 "\n"              \
		 , seq++, _src, _dst, _src_port, _src_port + 1, _dst_port, _dst_port + 1);


	len = strlen(buffer);
	w = 0;

	while(w != len) {
		tmp = write(_s, buffer + w, len - w);
		if (tmp == -1) {
			LOG(L_ERR, "close_pinhole(): %s\n", strerror(errno));
			return 0;
		}
		printf("Sent:\n%s", buffer);
			
		w += tmp;
	}
	DBG("close_pinhole(): %s \n", buffer);	
	DBG("close_pinhole(): %d bytes sent\n", w);

	return 1;
}



static int can_read(int _s, int _i)
{
	fd_set rfds;
	struct timeval tv;
	int retval;

	     /* Watch stdin (fd 0) to see when it has input. */
	FD_ZERO(&rfds);
	FD_SET(_s, &rfds);

	tv.tv_sec = _i;
	tv.tv_usec = 0;

	retval = select(_s + 1, &rfds, NULL, NULL, &tv);
	     /* Don't rely on the value of tv now! */

	if (retval) {
		return 1;
	} else {
		return 0;
	}
}


int read_reply(int _s, char* reply)
{
	int tmp, len;

	len = 0;

	while(can_read(_s, 2)) {
		tmp = read(_s, reply + len, 1);
		if (tmp == -1) {
			LOG(L_ERR, "read_reply(): %s\n", strerror(errno));
			printf("read_reply() error: \n");
			return 0;
		}
		len += tmp;
	}

	reply[len] = '\0';
	printf("rr:%s\n", reply);

	if (!len) {
		LOG(L_ERR, "read_reply(): No reply received within 2 seconds\n");
		return 0;
	}

	return 1;
}


int get_reply_code(void) 
{
	int res;

	sscanf(buffer, "FCP=1.0 %d", &res);
	return res;
}



int get_nat(int _s, char* _ip, int _p1, int _p2)
{
	char buffer[256];
	int len, w, tmp;

	if (!_ip) {
		LOG(L_ERR, "get_nat(): Invalid parameter value\n");
		return 0;
	}

	snprintf(buffer, 256, 
		 "QUERYNAT FCP=1.0 "                    \
		 "SEQ=%d \n"                              \
		 "PME SRCIP=%s; SRCPORT=%d-%d; PROTO=17 \n\n",
		 seq++, _ip, _p1, _p2);

	
	len = strlen(buffer);
	w = 0;

	while(w != len) {
		tmp = write(_s, buffer + w, len - w);
		if (tmp == -1) {
			LOG(L_ERR, "get_nat(): %s\n", strerror(errno));
			return 0;
		}

		w += tmp;
	}
	
	printf("Sent:\n%s", buffer);

	DBG("get_nat(): %d bytes sent\n", w);
	return 1;
}



static int read_line(int _s, char* buf)
{
	int len, tmp;

	len = 0;
	while(can_read(_s, 2)) {
		tmp = read(_s, buf + len, 1);
		if (tmp == -1) {
			LOG(L_ERR, "read_line(): %s\n", strerror(errno));
			return 0;
		}
		len += tmp;
		if (*(buf + len - tmp) == '\n') break;
	}

	*(buf + len) = '\0';
	return 1;
}


int read_nat_reply(int _s, char* _ip, int _p1, int _p2)
{
	char buffer[512];
	char* i;

	if ((!_ip) || (!_p1) || (!_p2)) {
		LOG(L_ERR, "read_nat_reply(): Invalid parameter value\n");
		return -1;
	}

	while(1) {
		if (!read_line(_s, buffer)) {
			LOG(L_ERR, "read_nat_reply(): Error while reading line\n");
			return -1;
		}
		
		if (strstr(buffer, "IP=")) {
			i = buffer + 3;
			while(*i != ' ') i++;
			memcpy(_ip, buffer + 3, i - (buffer + 3));
			_ip[i - (buffer + 3)] = '\0';
		
			i = strchr(i, '=');
			if (!i) {
				LOG(L_ERR, "read_nat_reply(): Can't find port\n");
				return -1;
			}
			i++;
			sscanf(i, "%d", &_p1);

			i = strchr(i, '=');
			if (!i) {
				LOG(L_ERR, "read_nat_reply(): Can't find port\n");
				return -1;
			}
			i++;
			sscanf(i, "%d", &_p2);
			return 1;
		}
	}	
}

