#include "../../str.h"
#include "../../parser/msg_parser.h"

#include "parser.h"
#include "text_ops.h"



/* 
   Gets the media info from the sdp message from the
   request and the response and puts them in
   a struct that will be used to query the firewall
*/

int parse_media_info(struct sip_msg *request, struct sip_msg *reply, struct media_info *media) {

  str sdp;

  /* get the SDP part of the response */
  
  sdp.s = get_body( reply );
  if (sdp.s==0) {
    LOG(L_ERR,"ERROR:parse_media_info cannot extract sdp from msg!\n");
    goto error;
  }
  sdp.len = (int)reply->content_length->parsed;

  /*
    Get the media information from the response.
    Starting with the SRC's IP address...
  */


  media->src_host.s=get_char_value(sdp.s, "IP4");

  if (media->src_host.s ==NULL) {
    printf("parse_media_info: Error getting source Port\n");
    goto error;
  }

  media->src_host.len = strlen(media->src_host.s);

  printf("parse_media_info: Src HOST = %s. Len = %d\n", media->src_host.s, media->src_host.len);


  /* Get the SRC's port */

  media->src_port=get_int_value(sdp.s, "audio");

  if (media->src_port ==-1) {
    printf("parse_media_info: Error getting source Port\n");
    goto error;
  }

  media->src_port_str.s=get_char_value(sdp.s, "audio");

  if (media->src_port_str.s==NULL) {
    printf("parse_media_info: Error getting source Port\n");
    goto error;
  }

  media->src_port_str.len = strlen(media->src_port_str.s);

  printf("parse_media_info: Src PORT = %s.Len = %d\n", media->src_port_str.s, media->src_port_str.len);
  

  /* Now get the SDP part of the original request */
  
  
  sdp.s = get_body( request );
  if (sdp.s==0) {
    LOG(L_ERR,"ERROR:parse_media_info cannot extract sdp from msg!\n");
    goto error;
  }
  sdp.len = (int)request->content_length->parsed;

  /*
    Get the media information from the request.
    Starting with the DST's IP address...
  */

  media->dst_host.s=get_char_value(sdp.s, "IP4");

  if (media->dst_host.s ==NULL) {
    printf("parse_media_info: Error getting dst Port\n");
    goto error;
  }

  media->dst_host.len = strlen(media->dst_host.s);

  printf("parse_media_info: Dst HOST = %s.Len = %d\n", media->dst_host.s, media->dst_host.len);


  /* Get the DST's port */


  media->dst_port=get_int_value(sdp.s, "audio");

  if (media->dst_port ==-1) {
    printf("parse_media_info: Error getting dst Port\n");
    goto error;
  }

  media->dst_port_str.s=get_char_value(sdp.s, "audio");

  if (media->dst_port_str.s==NULL) {
    printf("parse_media_info: Error getting source Port\n");
    goto error;
  }

  media->dst_port_str.len = strlen(media->dst_port_str.s);

  printf("parse_media_info: Dst PORT = %s.Len = %d\n", media->dst_port_str.s, media->dst_port_str.len);
 

  return 1; /* Successful */


 error:
  return -1;
}

/*
  Get the nat response and process it. Put useful
  info into rpl
*/

int parse_fcp_info(char *s, struct fcp_info *rpl, struct media_info *sent) {

  char *string_start, *buf;
  char *string_end, *ptr;
  int end, len;

  /* Information sent to FCP server */

  rpl->sent_host.s = sent->src_host.s;
  rpl->sent_host.len = strlen(sent->src_host.s);
  rpl->sent_rtp_port.s = sent->src_port_str.s;
  rpl->sent_rtp_port.len = strlen(sent->src_port_str.s);
  //  rpl->sent_rtcp_port.s = rpl->sent_rtp_port + 1;


  /* IP information from FCP server */

  end = get_offset_end(s, "IP");
  string_start = skip_delims(s + end);
  string_end = skip_chars(string_start);
	
  len = string_end - string_start;
  
  buf = pkg_malloc( len+1 ); /* extra character for \0? */
  if(!buf) {
    LOG(L_ERR,"ERROR:parse_fcp_info: out of memory! \n");
    goto error;
  }
  
  rpl->rpl_host.s = buf;
  memcpy( rpl->rpl_host.s, string_start , len);
  *(rpl->rpl_host.s + len) = '\0';
  rpl->rpl_host.len=len;
  printf("parse_fcp_info: IP = %s. Length: %d\n", rpl->rpl_host.s, len);


  /* Port information */

  end = get_offset_end(s, "PORT");
  string_start = skip_delims(s + end);
  string_end = skip_chars(string_start);
	
  len = string_end - string_start;
  
  buf = pkg_malloc( len+1 ); /* extra character for \0? */
  if(!buf) {
    LOG(L_ERR,"ERROR:parse_fcp_info: out of memory! \n");
    goto error;
  }
 

  rpl->rpl_rtp_port.s = buf;
  memcpy( rpl->rpl_rtp_port.s, string_start , len);
  *(rpl->rpl_rtp_port.s + len) = '\0';
  //  rpl->rpl_rtp_port = string_to_int(buf);
  //  rpl->rpl_rtcp_port = rpl->rpl_rtp_port+1;
  rpl->rpl_rtp_port.len = strlen(rpl->rpl_rtp_port.s);
  printf("parse_fcp_info: RTP PORT = %s\n", rpl->rpl_rtp_port.s);


  /* Seq information */

  end = get_offset_end(s, "SEQ");
  string_start = skip_delims(s + end);
  string_end = skip_chars(string_start);
	
  len = string_end - string_start;
  
  buf = pkg_malloc( len+1 ); /* extra character for \0? */
  if(!buf) {
    LOG(L_ERR,"ERROR:parse_fcp_info: out of memory! \n");
    goto error;
  }
 
  ptr = buf;
  memcpy( ptr, string_start , len);
  *(ptr + len) = '\0';
  rpl->seq = string_to_int(buf);
  printf("parse_fcp_info: SEQ = %d. Length: %d\n", rpl->seq, len);

  return 1; /* Successful */

 error:

  return -1;
}

/* 
   Checks whether <message> is of type <method>.
   <method> is a null terminated string.
   Returns 1 if <message> is of type <method>,
   0 if false.
*/

int is_method(char *message, char *method) {

  int len, offset;
  char *string_start;

  len = strlen(method);

  offset = get_offset_end(message, "SIP/2.0");
  string_start = skip_delims(message + offset);

  return (strncmp(string_start, method, (size_t)len)==0);

}
