#include <string.h>
#include <stdio.h>
#include <fcntl.h>

#include "text_ops.h"
#include "../../mem/mem.h"
#include "../../data_lump_rpl.h"



/*********************** match ***************************
 
   Match a substring <pattern> within a string <string>.
   It will return the offset to the beginning of <pattern>
   in pm->rm_so, and the offset to the end of <pattern>
   in pm->rm_eo. string + pm->rm_eo is the address to
   the beginning of <pattern>

**********************************************************/

int match(const char *string, char *pattern, regmatch_t *pm)
{
    int    status;
    regex_t    re;

    if (regcomp(&re, pattern, REG_EXTENDED) != 0) {
        return(-1);      /* Report error. */
    }
    status = regexec(&re, string, (size_t) 1, pm, 0);
    regfree(&re);
    if (status != 0) {
        return(-1);      /* Report error. */
    }
    return(0); /* Successful */
}

/* 
   Convert string of chars (NULL terminated) to integer
*/

long int string_to_int(char *s) {

  return strtol(s, (char **)NULL, 10);

}

/* 
   Convert int to char. Return number of
   characters printed. -1 if error ocurred
*/

int int_to_char(char *dest, int src, int size) {

  return snprintf(dest, size, "%d", src);

}

/*
  Returns the offset to the end of the first appearance
  of <key> within <string>
*/

int get_offset_end(char *string, char *key) {

  regmatch_t pmatch;
  int status;

  status = match(string, key, &pmatch);
  if (status != 0) {
    return(-1);      /* Report error. */
  } else return pmatch.rm_eo;
}

/*
  Returns the offset to the beginning of the first appearance
  of <key> within <string>
*/

int get_offset_start(char *string, char *key) {

  regmatch_t pmatch;
  int status;

  status = match(string, key, &pmatch);
  if (status != 0) {
    return(-1);      /* Report error. */
  } else return pmatch.rm_so;
}

/* 
   Jumps characters till next delim
*/

char* skip_chars(char* _s)
{
  while ((*_s != ' ') && (*_s != '\r') && \
	 (*_s != '\n') && (*_s != '\t') && \
	 (*_s != ';') && (*_s != '\0') && \
	 (*_s !='-') && (*_s !='=') && \
	 (*_s !='<') && (*_s !='>') && \
	 (*_s !=':')) _s++;
  return _s;
}

/*
  Jumps characters till next word
*/

char* skip_delims(char* _s) {
  while (((*_s == ' ') || (*_s == '\r') || \
	 (*_s == '\n') || (*_s == '\t') || \
	 (*_s == ';') || (*_s == '-') || \
	 (*_s =='<') || (*_s =='>') || \
	 (*_s ==':') ||	(*_s == '=')) && \
	 (*_s != '\0')) _s++;
  return _s;
}


/* 
   Gets the value pointed by <key> from a string <s> and
   parses it as char* to <dest>. Returns the value of <key>.
   It expects s to contain
   key<delim1>...<delimn>value, where <delim> is {":", ";", 
   " ", ">", "<", "-", "\n", "\t", "\r" }
*/
 
int get_int_value(char *s, char *key) {


  char *buf, *ptr;
  int len;
  int end;

  char *string_end;
  char *string_start;
  

  end = get_offset_end(s, key);
  if(end < 0) {
    printf("SRC Port not found\n");
    return -1;
  }
  string_start = skip_delims(s + end);
  string_end = skip_chars(string_start);
	
  len = string_end - string_start;
  
  buf = pkg_malloc( len+1 ); /* extra character for \0? */
  if(!buf) {
    LOG(L_ERR,"ERROR:extcmd_add_contact: out of memory! \n");
    return -1;
  }
  
  ptr = buf;
  memcpy( ptr, string_start , len);
  *(ptr + len) = '\0';
  return string_to_int(buf);

}

/* 
   Gets the value pointed by <key> from a string <s> and
   parses it as char* to <dest>. Returns the
   length of the value of <key>. It expects s to contain
   key<delim>...<delimn>value, where <delim> is {":", ";",
   " ", ">", "<", "-", "\n", "\t", "\r" }

*/
 
char* get_char_value(char *s, char *key) {

  char *buf;
  int len;
  int end;

  char *string_end;
  char *string_start;
  
  end = get_offset_end(s, key);
  if(end < 0) {
    printf("SRC IP not found\n");
    return NULL;
  }
  string_start = skip_delims(s + end);
  string_end = skip_chars(string_start);
	
  len = string_end - string_start;
  
  buf = pkg_malloc( len+1 ); /* extra character for \0? */
  if(!buf) {
    LOG(L_ERR,"ERROR:extcmd_add_contact: out of memory! \n");
    return NULL;
  }
  

  printf("get_char_value: len = %d\n", len);

  memcpy( buf, string_start , len);
  *(buf + len) = '\0';
  return buf;

}

